﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.WebSockets;
using System.ServiceModel.Syndication;
using System.Threading.Tasks;
using System.Timers;
using System.Xml;
using Discord;
using Discord.WebSocket;
using Google.Apis.Auth.OAuth2;
using Google.Apis.Sheets.v4;

namespace DiscoBot.gsmeet
{
    public class GSMeet : IModule
    {
        string IModule.Name { get => "GSMeet"; set => throw new NotImplementedException(); }
        private SocketGuild guild;
        private GSMeetContext gsmeetContext;
        private Dictionary<string, Timer> timers = new Dictionary<string, Timer>();
        private List<WebSocket> webSockets = new List<WebSocket>();

        public Dictionary<string, Func<SocketMessage, string[], Task>> Commands { get; set; } = new Dictionary<string, Func<SocketMessage, string[], Task>>();

        private static readonly string[] Scopes = { SheetsService.Scope.Spreadsheets };
        private static readonly string ApplicationName = "DiscoBot";
        private SheetsService service;

        public GSMeet(SocketGuild guild)
        {
            this.guild = guild;
            gsmeetContext = new GSMeetContext(guild.Id);
            gsmeetContext.Database.EnsureCreated();

            GoogleCredential credential;
            using(var stream = new FileStream("client_secret.json", FileMode.Open, FileAccess.Read))
            {
                credential = GoogleCredential.FromStream(stream).CreateScoped(Scopes);
            }
            service = new SheetsService(new Google.Apis.Services.BaseClientService.Initializer()
            {
                HttpClientInitializer = credential,
                ApplicationName = ApplicationName
            });

            foreach(var f in gsmeetContext.GSheets)
            {
                InitializeSheet(f);
            }

            Commands.Add("gsmeetadd", HandleGSMeetAddCommand);
            Commands.Add("gsmeetdel", HandleGSMeetDelCommand);
            Commands.Add("gsmeetlist", HandleGSMeetListCommand);
            Commands.Add("gsmeetdebug", HandleGSMeetDebugCommand);
        }

        private Task InitializeSheet(GSheet sheet)
        {
            Console.WriteLine("Found sheet " + sheet.Name);
            Timer timer = new Timer(sheet.CheckInterval.TotalMilliseconds);
            timer.AutoReset = true;
            timer.Elapsed += async (sender, e) =>
            {
                await Task.Run(() => HandleSheetCheck(sheet));
            };
            timer.Start();
            timers.Add(sheet.Name, timer);
            return Task.CompletedTask;
        }

        private Task DeinitializeSheets(GSheet sheet)
        {
            /*
            Timer t = timers[feed.Name];
            t.Stop();
            timers.Remove(feed.Name);
            */
            return Task.CompletedTask;
        }

        private Task HandleSheetCheck(GSheet sheet)
        {
            SocketTextChannel c = guild.Channels.Where(g => g.Id == sheet.Channel).Single() as SocketTextChannel;
            c.SendMessageAsync("Checking sheet " + sheet.SheetId + " :3");
            var values = FetchRangeFromSheet(sheet.SheetId, $"{sheet.SheetName}!A1:I10");
            if (values == null || values.Count < 1)
            {
                c.SendMessageAsync("No values found.");
                return Task.CompletedTask;
            }
            var str = "";
            foreach (var row in values)
            {
                foreach (var col in row)
                {
                    str += col + "| ";
                }
                str += "\n";
            }
            c.SendMessageAsync("Result: " + str);
            return Task.CompletedTask;
        }

        private Task HandleGSMeetAddCommand(SocketMessage msg, string[] parameters)
        {
            var gchan = msg.Channel as IGuildChannel;
            string name = parameters[1];
            string sheetId = parameters[2];
            string sheetName = parameters[3];
            try
            {
                int timeSec = Int32.Parse(parameters[4]);
                GSheet sheet = new GSheet();
                sheet.Channel = gchan.Id;
                sheet.Name = name;
                sheet.SheetId = sheetId;
                sheet.SheetName = sheetName;
                sheet.LastChecked = DateTimeOffset.Now;
                sheet.CheckInterval = new TimeSpan(0, 0, timeSec);
                gsmeetContext.GSheets.Add(sheet);
                try
                {
                    gsmeetContext.SaveChanges();
                    InitializeSheet(sheet);
                    msg.Channel.SendMessageAsync("Sheet " + sheet.Name + " with id " + sheet.SheetId + " saved.");
                }
                catch (InvalidOperationException)
                {
                    msg.Channel.SendMessageAsync("Unable to save sheet feed.");
                }
            }
            catch (FormatException)
            {
                msg.Channel.SendMessageAsync("Unable to save sheet. Invalid check time.");
            }
            return Task.CompletedTask;
        }

        private Task HandleGSMeetListCommand(SocketMessage msg, string[] parameters)
        {
            /*
            List<RssFeed> feeds = rssContext.RssFeeds.Where(f => f.Channel == msg.Channel.Id).ToList();
            List<string> m = new List<string>();
            m.Add("Feeds for this channel are:");
            m.Add("Name | URL | CheckInterval | LastChecked");
            msg.Channel.SendMessageAsync(string.Join("\n", m));
            m = new List<string>();
            int cnt = 0;
            foreach (var f in feeds)
            {
                m.Add("- " + f.Name + " | " + f.Url + " | " + f.CheckInterval + " | " + f.LastChecked);
                cnt++;
                if(cnt > 4)
                {
                    cnt = 0;
                    msg.Channel.SendMessageAsync(string.Join("\n", m));
                }
            }
            if(cnt > 0)
                msg.Channel.SendMessageAsync(string.Join("\n", m));
                */
            return Task.CompletedTask;
        }

        private Task HandleGSMeetDelCommand(SocketMessage msg, string[] parameters)
        {
            /*
            var gchan = msg.Channel as IGuildChannel;
            string name = parameters[1];
            try
            {
                RssFeed f = rssContext.RssFeeds.Where(f => f.Name == name).Single();
                rssContext.Remove(f);
                rssContext.SaveChanges();
                DeinitializeFeed(f);
                msg.Channel.SendMessageAsync("Removed feed " + f.Name);
            } catch(InvalidOperationException)
            {
                msg.Channel.SendMessageAsync("Could not find feed " + name);
            }
            */
            return Task.CompletedTask;
        }

        private Task HandleGSMeetDebugCommand(SocketMessage msg, string[] parameters)
        {
            var gchan = msg.Channel as IGuildChannel;
            string name = parameters[1];
            string sheetId = parameters[2];
            string sheetName = parameters[3];

            var values = FetchRangeFromSheet(sheetId, $"{sheetName}!A1:I10");
            if(values == null || values.Count < 1)
            {
                msg.Channel.SendMessageAsync("No values found.");
                return Task.CompletedTask;
            }
            var str = "";
            foreach(var row in values)
            {
                foreach(var col in row)
                {
                    str += col + "| ";
                }
                str += "\n";
            }
            msg.Channel.SendMessageAsync("Result: "+str);

            return Task.CompletedTask;
        }

        private IList<IList<object>> FetchRangeFromSheet(string sheetId, string range)
        {
            var request = service.Spreadsheets.Values.Get(sheetId, range);
            var response = request.Execute();
            return response.Values;
        }

        public void Initialize()
        {
            Console.WriteLine("Initializing gsmeet...");
        }

        public async Task OnNewWebSocketAsync(WebSocket ws, TaskCompletionSource<object> tcs)
        {
            Console.WriteLine("Calendar " + guild.Id + " has a new websocket.");
            webSockets.Add(ws);
        }
    }
}
